% Copyright (c) Camden Dixie O'Brien
% SPDX-License-Identifier: AGPL-3.0-only

-module(tcp_server).
-behaviour(gen_server).

-export([start_link/1, child_spec/1]).
-export([init/1, handle_call/3, handle_cast/2, handle_info/2,
		 terminate/2, code_change/3]).

-record(state, {socket, acceptor}).

start_link(Port) ->
	gen_server:start_link({local, ?MODULE}, ?MODULE, [Port], []).

child_spec(Port) ->
	#{id => ?MODULE,
	  start => {?MODULE, start_link, [Port]},
	  restart => permanent,
	  shutdown => 5000,
	  type => worker,
	  modules => [?MODULE]}.

init([Port]) ->
	{ok, Socket} = gen_tcp:listen(Port, [binary, {active, true}]),
	Pid = spawn_link(fun() -> acceptor_loop(Socket) end),
	{ok, #state{socket = Socket, acceptor = Pid}}.

handle_call(_Request, _From, State) ->
	{reply, ok, State}.

handle_cast(_Msg, State) ->
	{noreply, State}.

handle_info({'EXIT', Pid, Reason}, State = #state{acceptor = Pid}) ->
	{stop, {acceptor_died, Reason}, State};
handle_info(_Info, State) ->
	{noreply, State}.

terminate(_Reason, #state{socket = Socket}) ->
	gen_tcp:close(Socket),
	ok.

code_change(_OldVsn, State, _Extra) ->
	{ok, State}.

acceptor_loop(Socket) ->
	case gen_tcp:accept(Socket) of
		{ok, ClientSocket} ->
			gen_tcp:controlling_process(
			  ClientSocket,
			  spawn(fun() -> handle_connection(ClientSocket) end)),
			acceptor_loop(Socket);
		{error, closed} ->
			ok
	end.

handle_connection(Socket) ->
	receive
		{tcp, Socket, Data} ->
			gen_tcp:send(Socket, Data),
			handle_connection(Socket);
		{tcp_closed, Socket} ->
			ok
	end.
